
/*
 * $Id: Client.cc,v 1.43.2.32 1999/12/26 23:38:17 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <iomanip.h>

#include "CltConnMgr.h"
#include "CltOpts.h"
#include "CltXact.h"
#include "EphPortMgr.h"
#include "ExpPortMgr.h"
#include "OidCache.h"
#include "PopModel.h"
#include "RndPermut.h"
#include "ServerRep.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "LogComment.h"
#include "ErrorMgr.h"
#include "CltCfg.h"
#include "Client.h"

#include "RobotSym.h"

#include "polyBcastChannels.h"
#include "polyLogTags.h"
#include "polyErrors.h"
#include "globals.h"

struct SharedCfgItem { const RobotSym *sym; CltCfg *cfg; };
static Array<SharedCfgItem> SharedCfgs;

ObjFarm<CltXact> Client::TheXacts;
Array<PortMgr*> Client::ThePortMgrs;
Array<ServerRep*> Client::TheServers;
Ring<int> *Client::TheWarmupPlan = 0;

int Client::TheClientCnt = 0;
int Client::TheShrdWorkSetCap = -1;


Client::Client(): theCfg(0), theConnMgr(0), theCache(0),
	theCcXactLvl(0), theLaunchCount(0), theExtraLaunchLvl(0) {

	theChannels.append(TheInfoChannel);
	theChannels.append(TheLogCfgChannel);
	theChannels.append(TheLogStateChannel);
	startListen();
}

Client::~Client() {
}

void Client::configure(const RobotSym *cfg, const NetAddr &aHost, bool bind2iface) {
	SockOpt opt;
	Agent::configure(cfg, aHost, opt);
	theHost.port(-1); // clients do not care about port number

	theCfg = sharedConfig(cfg);

	thePrivOids.stretch(theCfg->theSrvCount);
	thePrivWorkSetCap.stretch(theCfg->theSrvCount);
	for (int i = 0; i < theCfg->theSrvCount; ++i) {
		thePrivOids.append(0);
		thePrivWorkSetCap.append(-1);
	}

	int openConnLmt = -1;
	cfg->openConnLimit(openConnLmt);
	theConnMgr = new CltConnMgr;
	theConnMgr->configure(opt, cfg->pconnUseLmt(), openConnLmt,
		TheCltOpts.thePxyAddr ? 1 : theCfg->theSrvCount);
	theConnMgr->portMgr(getPortMgr(bind2iface));
	theConnMgr->idleTimeout(cfg->idlePconnTimeout());

	int cacheCap = 0;
	if (cfg->cacheCap(cacheCap))
		theCache = new OidCache(cacheCap);

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Client::start() {
	Assert(!theLaunchCount);
	Assert(theConnMgr);

	Agent::start();

	// first client that gets here will lock the phase and build the plan
	if (!TheClientCnt) {
		TheStatPhaseMgr->lock();
		BuildWarmupPlan();
	}

	TheClientCnt++;

	if (theCfg->theLaunchWin > 0) {
		// uniformly distributed start delay
		static RndGen rng;
		const double delay = rng(0.0, theCfg->theLaunchWin.secd());
		sleepFor(Time::Secd(delay));
	} else {
		scheduleLaunch(TheClock);
	}
}

void Client::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWorkSetFreeze) {
		for (int i = 0; i < thePrivWorkSetCap.count(); ++i)
			thePrivWorkSetCap[i] = thePrivOids[i];

		if (TheShrdWorkSetCap < 0)
			FreezeSrvWorkSetCap();
	}
}

void Client::noteLogEvent(BcastChannel *ch, OLog &log) {
	if (ch == TheLogCfgChannel) {
		log << bege(lgCltCfg);
		// XXX: implement (CltCfgRec)
		log << ende;
	} else
	if (ch == TheLogStateChannel) {
		log << bege(lgCltState);
		log << theSeqId 
			<< theLaunchCount
			<< thePrivOids << thePrivWorkSetCap;
		log << ende;
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);

	// cache cachable embedded objects
	if (theCache) {
		const OidRec &oidr = x->oidRec();
		if (oidr.embedded() && oidr.cachable && oidr.oid > 0)
			theCache->putOid(oidr);
	}

	theConnMgr->put(x->conn());
	TheXacts.put(x);
	theCcXactLvl--;
	Assert(theCcXactLvl >= 0);

	// push waiting xactions forward
	if (theLaunchDebts.count() && !theConnMgr->atHardConnLimit())
		resumeXact();
}

void Client::scheduleLaunch(Time lastLaunch) {
	Assert(!thePendAlarmCnt);
	if (theThinkDistr) {
		const double lf = TheStatPhaseMgr->loadFactor(lastLaunch, theLaunchCount);
		Assert(lf >= 0);

		if (lf <= 0) // zero load: no more requests needed
			return;

		double delay;
		if (TheStatPhaseMgr->varyLoad()) {
			// next launch should happen [0, 2dt] from last launch
			// where dt is 1/current_rate
			static RndGen rng;
			delay = rng(0.0, 2.*theThinkDistr->mean()/lf);
		} else {
			delay = theThinkDistr->trial()/lf;
		}
		//cerr << here << TheClock << "| " << seqId() << " delay: " << delay << " lf: " << lf << endl;
		sleepTill(lastLaunch + Time::Secd(delay));
	} else {
		tryLaunch();
	}
}

// will create a xaction if needed
// must be ready to launch a xaction
bool Client::launch(CltXact *x) {
	Assert(!theConnMgr->atHardConnLimit());

	if (!x)
		x = genXact();

	// connect and execute the transaction
	const NetAddr &connAddr = TheCltOpts.thePxyAddr ?
		(const NetAddr &)TheCltOpts.thePxyAddr : 
		(const NetAddr &)(GetServerRep(x->oidRec().sidx)->addr());

	if (Connection *conn = theConnMgr->get(connAddr)) {
		theLaunchCount++;
		theCcXactLvl++;
		x->exec(this, conn);
		return true;
	}

	if (ReportError(errno))
		Comment(1) << theHost << " failed to connect to " << connAddr << endc;
	Broadcast(TheXactErrChannel, x);

	TheXacts.put(x);

	// if no alarms are pending, we may get stuck and submit
	// no more xactions (e.g. best effort); this is a kludge
	if (!thePendAlarmCnt)
		sleepTill(TheClock); // submit more next cycle

	return false;
}

// will launch a xaction (if possible) or postpone it (if not)
bool Client::tryLaunch(CltXact *x) {
	// check if we should postpone the xaction
	if (theConnMgr->atHardConnLimit()) {
		if (theCfg->theWaitXactLmt < 0 || theLaunchDebts.count() < theCfg->theWaitXactLmt) {
			suspendXact(x);
			return true;
		}

		if (ReportError(errTooManyWaitXact))
			Comment(3) << "xactions active: " << theCcXactLvl 
				<< " waiting: " << theLaunchDebts.count() 
				<< " limit: " << theCfg->theWaitXactLmt << endc;
		if (x)
			TheXacts.put(x);
		return false;
	}

	return launch(x);
}

void Client::suspendXact(CltXact *x) {
	theLaunchDebts.append(x);
	Broadcast(TheWaitBegChannel, x);
}

void Client::resumeXact() {
	CltXact *x = theLaunchDebts.pop();
	Broadcast(TheWaitEndChannel, x);
	launch(x);
}

CltXact *Client::genXact(const OidRec &or) {
	CltXact *x = TheXacts.get();
	x->oidRec(or);
	return x;
}

CltXact *Client::genXact() {
	OidRec or;

	// at the beginning, touch all servers so we have
	// private or secured id for each server to generate hits
	if (TheWarmupPlan)
		or.sidx = genWarmupSrvIdx();

	if (or.sidx < 0)
		or.sidx = (int)theCfg->theOriginSel->trial();

	genOid(or);
	or.tid = 0; // direct access
	return genXact(or);
}

void Client::genOid2(OidRec &or) {
	ServerRep &srvRep = *GetServerRep(or.sidx);
	//cerr << here << "pid: " << theId << " last priv oid: " << thePrivOids[or.sidx] << " server: " << or.sidx << " wid: " << srvRep.id() << " sec oid: " << srvRep.secOid() << endl;

	const bool forcePubOid = srvRep.secOid() <= 0 && srvRep.haveNewOids();

	static RndGen rng;
	const bool genPubOid = forcePubOid || rng.event(theCfg->thePublicRatio);
	const bool repeatOid = rng.event(theCfg->theRecurRatio*TheStatPhaseMgr->recurFactor());

	/* this logic gives priority to repeatOid goal rather than genPubOid */

	if (repeatOid) {
		or.repeat = true;

		const int privOid = thePrivOids[or.sidx];
		const int pubOid = srvRep.secOid();

		if (pubOid && (genPubOid || !privOid)) {
			or.wid = srvRep.id(); // use server id
			or.oid = theCfg->thePopModel->choose(rng, pubOid, srvRep.workSetCap());
			return;
		}
		if (privOid) {
			or.wid = theId; // use this robot id
			or.oid = theCfg->thePopModel->choose(rng, privOid, thePrivWorkSetCap[or.sidx]);
			return;
		}
		// failed to repeat, fall through
	}

	or.repeat = false;
	if (genPubOid) {
		// can ask for a new oid?
		if (srvRep.haveNewOids()) {
			or.wid = srvRep.id(); // use server id
			or.oid = srvRep.newOid();
			return;
		}

		// report insufficient number of new oids if server is initialized
		if (srvRep.secOid() > srvRep.maxNewOidCap()) {
			ReportError(errNoNewOids);
			srvRep.noteNoNewOids();
		}
		// failed to produce new public oid, fall through
	}
		
	// new private object [last resort, never fails]
	or.wid = theId; // use this robot id
	or.oid = ++thePrivOids[or.sidx];
}

void Client::genOid(OidRec &or) {
	genOid2(or);
	if (theCfg->genUniqUrls)
		or.wid = UniqId::Create(); // changes every time
}

void Client::noteEmbedded(CltXact *, const OidRec &or) {
	static RndGen rng;
	if (!rng.event(theCfg->theEmbedRecurRatio))
		return;

	if (theCache && theCache->hasOid(or))
		return; // XXX: count these

	if (tryLaunch(genXact(or)))
		theExtraLaunchLvl++;
}

void Client::scanServers(const RobotSym *rs, CltCfg *cfg) {
	Array<NetAddr*> addrs;
	if (!rs->origins(addrs, cfg->theOriginSel) || !addrs.count()) {
		Comment(1) << "no origin addresses specified for robot " << rs->kind() << endc;
		FatalError(errOther);
	}
	Assert(cfg->theOriginSel);
	cfg->theOriginSel->rndGen(GenRndGen("client_origins"));
	cfg->theSrvCount = addrs.count();

	const int savedCount = TheServers.count();
	for (int i = 0; i < addrs.count(); ++i)
		cAddServer(*addrs[i], !savedCount);

	if (savedCount && TheServers.count() != savedCount) {
		Comment(1) << "varying the number of origin servers per robot "
			<< "is not yet supported for memory consumption reasons; "
			<< "let us know if you need this feature" << endl;
		FatalError(errOther);
	}

	// get rid of addresses
	while (addrs.count()) delete addrs.pop();
}

PortMgr *Client::getPortMgr(bool bind2iface) {
	const NetAddr addr = bind2iface ? theHost : NetAddr();

	// check if we already have a port mgr for host
	for (int i = 0; i < ThePortMgrs.count(); ++i) {
		if (ThePortMgrs[i]->addr() == addr)
			return ThePortMgrs[i];
	}

	// create new port manager
	PortMgr *mgr = TheCltOpts.thePorts.set() ?
		(PortMgr*) new ExpPortMgr(addr, TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		(PortMgr*) new EphPortMgr(addr);
	ThePortMgrs.append(mgr);
	return mgr;
}

bool Client::srvReadyToHit(const ServerRep &s) const {
	// already got secured oid OR
	// was accessed and will never get a secured oid
	return s.secOid() > 0 || (theCfg->genUniqUrls && s.accesses() > 0);
}

int Client::genWarmupSrvIdx() {
	Assert(TheWarmupPlan);
	int idx = -1;

	static int touchCnt = 0;
	while (!TheWarmupPlan->empty() && idx < 0) {
		const int i = TheWarmupPlan->dequeue();

		if (srvReadyToHit(*TheServers[i]))
			continue;

		TheWarmupPlan->enqueue(i);
		idx = i;
		touchCnt++;
	}

	if (TheWarmupPlan->count()) {
		// once in a while, report progress
		const Time reportGap = Time::Sec(45); // how often to report
		static const Time start = TheClock;
		static Time nextReport = start + reportGap;
		Clock::Update(false);
		if (nextReport <= TheClock) {

			int readyCnt = 0;
			for (int i = 0; i < TheServers.count(); ++i) {
				if (srvReadyToHit(*TheServers[i]))
					readyCnt++;
			}

			const double touchp = Percent(touchCnt, 2*TheServers.count());
			const double readyp = Percent(readyCnt, TheServers.count());
			const double progrp = MiniMax(touchp, readyp, 99.0);
			Comment(5) << "fyi: server scan is probably "
				<< setw(6) << progrp << "% completed with " 
				<< setw(5) << readyCnt << " out of "
				<< TheServers.count() << " servers ("
				<< setw(6) << Percent(readyCnt, TheServers.count())
				<< "%) ready to be hit" << endc;
			nextReport = TheClock + reportGap;
		}
	} else {
		delete TheWarmupPlan;
		TheWarmupPlan = 0;

		Comment(3) << "fyi: server scan completed with all " << TheClientCnt << " local robots "
			<< "ready to hit all " << TheServers.count() << " servers" << endc;
		TheStatPhaseMgr->unlock();
	}
	
	return idx;
}

// creates new server representative if needed
// returns true iff creation was needed
bool Client::cAddServer(const NetAddr &addr, bool forceAdd) {
	if (!forceAdd) {
		// check if we already know about this server
		for (int i = 0; i < TheServers.count(); ++i) {
			if (TheServers[i]->addr() == addr)
				return false;
		}
	}
	
	// new server
	TheServers.append(new ServerRep(addr));
	return true;
}

ServerRep *Client::GetServerRep(int idx) {
	Assert(0 <= idx && idx < TheServers.count());
	return TheServers[idx];
}


CltCfg *Client::sharedConfig(const RobotSym *rs) {
	CltCfg *cfg =0;
	for (int i = 0; i < SharedCfgs.count(); ++i) {
		if (SharedCfgs[i].sym == rs) {
			cfg = SharedCfgs[i].cfg;
			break;
		}
	}

	if (!cfg) {
		cfg = new CltCfg;
		cfg->configure(rs);
		SharedCfgItem ci;
		ci.sym = rs;
		ci.cfg = cfg;
		SharedCfgs.append(ci);
		// do this once per shared configuration
		scanServers(rs, cfg);
	}

	return cfg;
}

void Client::FreezeSrvWorkSetCap() {
	Assert(TheShrdWorkSetCap < 0);

	TheShrdWorkSetCap = 0;
	for (int i = 0; i < TheServers.count(); ++i) {
		TheServers[i]->freezeWorkSetCap();
		TheShrdWorkSetCap += TheServers[i]->workSetCap();
	}
}

void Client::BuildWarmupPlan() {
	Assert(!TheWarmupPlan);

	TheWarmupPlan = new Ring<int>(TheServers.count());

	// populate with server indeces
	for (int idx = 0; idx < TheServers.count(); ++idx)
		TheWarmupPlan->enqueue(idx);

	TheWarmupPlan->randomize();
}

static
OLog &operator <<(OLog &log, ServerRep *sr) {
	sr->store(log);
	return log;
}

void Client::LogState(OLog &log) {
	log << bege(lgSrvRepState) << TheServers << ende;
}
