
/*
 * $Id: EmbedContMdl.cc,v 1.4.2.4 1999/10/26 16:18:01 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Farm.h"
#include "ContBuildIter.h"
#include "ContentCfg.h"
#include "ContentMgr.h"

#include "ContentSym.h"

#include "EmbedContMdl.h"
#include "rndDistrStat.h"


static ObjFarm<ContBuildIter> TheIterators;


EmbedContMdl::EmbedContMdl(): theCfgSel(0), theEmbedCount(0),
	theEmbedCountMean(-1) {
	TheIterators.limit(1024); // magic, no good way to estimate
}


void EmbedContMdl::configure(const ContentSym *cfg) {
	Assert(!theEmbedCount);

	if ((theEmbedCount = cfg->embedCount()))
		theEmbedCountMean = RndDistrStat(theEmbedCount).mean();

	configureEmbedCfgs(cfg);

	if (theEmbedCount && !theCfgs.count()) {
		cerr << cfg->loc() 
			<< "the number of embedded objects per container is specified, "
			<< "but object content types are undefined"
			<< endl;
		exit(-2);
	}

	if (!theEmbedCount && theCfgs.count()) {
		cerr << cfg->loc() 
			<< "object content types are specified, "
			<< "but the number of embedded objects per container is not"
			<< endl;
		exit(-2);
	}
}

double EmbedContMdl::compContPerCall(const ContentCfg *cc) const {
	int idx = 0;
	if (find(cc->id(), idx))
		return theEmbedCountMean * theProbs[idx];
	return 0.0;
}

void EmbedContMdl::noteNewContProb(ContentCfg *cc, double newProb) {
	int idx = 0;
	if (!find(cc->id(), idx))
		return;

	theNewContProb[idx] = newProb;
	Assert(cc->newPerOid() > 0);
}

ContBuildIter *EmbedContMdl::getIterator(Size contSize, WrBuf *buf) {
	ContBuildIter *i = TheIterators.get();
	i->configure(this, buf, contSize);
	return i;
}

void EmbedContMdl::putIterator(ContBuildIter *i) {
	TheIterators.put(i);
}

int EmbedContMdl::embedGoal(RndGen &rng) {
	if (!theEmbedCount)
		return 0;
	theEmbedCount->rndGen(&rng); // note: we do not reset rng back
	return (int)theEmbedCount->trial();
}

int EmbedContMdl::embedOid(RndGen &rng, int oid, int &embedType, int &newOidCount) {
	Assert(theEmbedCount);

	// first select the embeded type
	theCfgSel->rndGen(&rng);   // note: we do not reset rng back
	const int idx = (int)theCfgSel->trial();
	Assert(0 <= idx && idx < theCfgs.count());
	const ContentCfg *cc = theCfgs[idx];
	embedType = cc->id();

	newOidCount++;
	Assert(newOidCount <= 255); // to simplify oid generation scheme (XXX)
	return abs((oid << 8) | newOidCount);
}

void EmbedContMdl::configureEmbedCfgs(const ContentSym *cfg) {
	Array<ContentSym*> syms;
	cfg->contains(syms, theCfgSel, theProbs);

	theCfgs.stretch(syms.count());
	for (int i = 0; i < syms.count(); ++i) {
		const ContentCfg *cfg = TheContentMgr.get(syms[i]);
		theCfgs.append(cfg);
	}

	theNewContProb.stretch(syms.count());
	theNewContProb.count(syms.count());
}

bool EmbedContMdl::find(int cfgId, int &idx) const {
	for (idx = 0; idx < theCfgs.count(); ++idx) {
		if (theCfgs[idx]->id() == cfgId)
			return true;
	}
	return false;
}
