
/*
 * $Id: StatIntvlRec.cc,v 1.8.6.9 1999/12/25 23:36:10 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <math.h>

#include "ILog.h"
#include "OLog.h"
#include "StatIntvlRec.h"

#include "gadgets.h"


StatIntvlRec::StatIntvlRec() {
	restart();
}

void StatIntvlRec::restart() {
	theWaitLvl.restart();
	theXactLvl.restart();
	theOpenLvl.restart();
	theEstbLvl.restart();

	theConnLifeTm.reset();
	theConnUseCnt.reset();

	theIdealHR.reset();
	theRealHR.reset();
	theChbR.reset();

	theHitRespTm.reset();
	theMissRespTm.reset();

	theFill.reset();
	theXactCnt = theXactErrCnt = 0;

	theScanLen.reset();

	theDuration = Time();
}

OLog &StatIntvlRec::store(OLog &log) const {
	Assert(theDuration >= 0);
	return log
		<< theWaitLvl << theXactLvl << theOpenLvl << theEstbLvl
		<< theConnLifeTm << theConnUseCnt
		<< theIdealHR << theRealHR << theChbR
		<< theHitRespTm << theMissRespTm
		<< theFill
		<< theXactCnt
		<< theXactErrCnt
		<< theDuration
		<< theScanLen
		;
}

ILog &StatIntvlRec::load(ILog &log) {
	return log
		>> theWaitLvl >> theXactLvl >> theOpenLvl >> theEstbLvl
		>> theConnLifeTm >> theConnUseCnt
		>> theIdealHR >> theRealHR >> theChbR
		>> theHitRespTm >> theMissRespTm
		>> theFill
		>> theXactCnt
		>> theXactErrCnt
		>> theDuration
		>> theScanLen
		;
}

double StatIntvlRec::errRatio() const {
	return Ratio(theXactErrCnt, complCnt() + theXactErrCnt);
}

double StatIntvlRec::errPercent() const {
	return Percent(theXactErrCnt, complCnt() + theXactErrCnt);
}

// XXX: we should have Rate type, not double
double StatIntvlRec::reqRate() const {
	return theDuration > 0 ?
		Ratio(theXactLvl.incCnt(), theDuration.secd()) : -1;
}

double StatIntvlRec::repRate() const {
	return theDuration > 0 ?
		Ratio(theXactLvl.decCnt(), theDuration.secd()) : -1;
}

AggrStat StatIntvlRec::respTime() const {
	AggrStat h(theHitRespTm);
	h += theMissRespTm;
	return h;
}

BigSize StatIntvlRec::repSz() const {
	return theRealHR.xacts().size(); // hits + misses
}

Size StatIntvlRec::repSzMean() const {
	if (const int x = theRealHR.xacts().count())
		return (theRealHR.xacts().size()/x).byte();
	return -1;
}

Time StatIntvlRec::scanTime() const {
	return theDuration > 0 && theScanLen.count() ?
		theDuration / theScanLen.count() : Time();
}

Time StatIntvlRec::scanStepTime() const {
	return theDuration > 0 && theScanLen.sum() > 0 ?
		theDuration / theScanLen.sum() : Time();
}

void StatIntvlRec::keepLevels(const StatIntvlRec &prevIntvl) {
	theWaitLvl.keepLevel(prevIntvl.theWaitLvl);
	theXactLvl.keepLevel(prevIntvl.theXactLvl);
	theOpenLvl.keepLevel(prevIntvl.theOpenLvl);
	theEstbLvl.keepLevel(prevIntvl.theEstbLvl);
}

void StatIntvlRec::concat(const StatIntvlRec &r) {
	if (theDuration >= 0)
		theDuration += r.theDuration;
	join(r);
}

void StatIntvlRec::merge(const StatIntvlRec &s) {
	join(s);
}

void StatIntvlRec::join(const StatIntvlRec &s) {
	Assert(s.theDuration >= 0);
	if (theDuration < 0)
		theDuration = s.theDuration;

	theWaitLvl.mergeWith(s.theWaitLvl);
	theXactLvl.mergeWith(s.theXactLvl);
	theOpenLvl.mergeWith(s.theOpenLvl);
	theEstbLvl.mergeWith(s.theEstbLvl);

	theConnLifeTm += s.theConnLifeTm;
	theConnUseCnt += s.theConnUseCnt;

	theIdealHR += s.theIdealHR;
	theRealHR += s.theRealHR;
	theChbR += s.theChbR;

	theHitRespTm += s.theHitRespTm;
	theMissRespTm += s.theMissRespTm;
	
	theFill += s.theFill;
	theXactCnt += s.theXactCnt;
	theXactErrCnt += s.theXactErrCnt;

	theScanLen += s.theScanLen;
}

ostream &StatIntvlRec::print(ostream &os, const String &pfx) const {
	os << pfx << "req.rate:\t " << reqRate() << endl;
	os << pfx << "rep.rate:\t " << repRate() << endl;

	theIdealHR.print(os, "hit", "miss", pfx + "offered.");
	theRealHR.print(os, "hit", "miss", pfx);
	theChbR.print(os, "cachable", "uncachable", pfx);
	theChbR.xacts().print(os, pfx + "rep.");
	respTime().print(os, pfx + "rep.rptm.");
	theHitRespTm.print(os, pfx + "hit.rptm.");
	theMissRespTm.print(os, pfx + "miss.rptm.");
	theFill.print(os, pfx + "fill.");
	theXactLvl.print(os, pfx + "xact.");

	os << pfx << "err_xact.ratio:\t " << errPercent() << endl;
	os << pfx << "err_xact.count:\t " << theXactErrCnt << endl;
	os << pfx << "ok_xact.count: \t " << theXactCnt << endl;

	theWaitLvl.print(os, pfx + "wait.");
	theOpenLvl.print(os, pfx + "conn.open.");
	theEstbLvl.print(os, pfx + "conn.estb.");
	theConnLifeTm.print(os, pfx + "conn.ttl.");
	theConnUseCnt.print(os, pfx + "conn.use.");

	theScanLen.print(os, pfx + "so_scan.length.");
	os << pfx << "so_scan.rptm:     \t " << scanTime().msec() << endl;
	os << pfx << "so_scan.step.rptm:\t " << scanStepTime().msec() << endl;

	os << pfx << "duration:\t " << theDuration.secd() << endl;

	return os;
}
