
/*
 * $Id: lx.cc,v 1.23.4.13 1999/12/26 21:38:36 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "CmdLine.h"
#include "opts.h"

#include "ErrorStat.h"
#include "ErrorStrTbl.h"
#include "StatPhaseRec.h"
#include "LogIter.h"

#include "matchAndPrint.h"
#include "polyLogTags.h"


class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",         "list of options"),
			theOutFileName(this, "out <file>",   "redirect console output", "-"),
			thePhases(this,      "phases <list>","names of phase to extract"),
			theObjects(this,     "objects <list>","names of objects to extract")
			{}

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		StrOpt theOutFileName;
		StrArrOpt thePhases;
		StrArrOpt theObjects;
		Array<String*> theFiles;
};

struct PhaseGroup {
	static ErrorStrTbl TheErrStrs;

	ErrorStat errors;
	StatPhaseRec rec;
	String name;

	PhaseGroup(const String &aName);
};

class MyScanner {
	protected:
		typedef const Array<String*> &Strs;

	public:
		MyScanner(Strs phaseNames, Strs objectNames);
		~MyScanner();

		bool load(ILog &il, const LogEntryPx &px);

		void report() const;
		
	protected:
		void notePhase(const StatPhaseRec &ph);

	protected:
		Strs thePhaseNames;
		Strs theObjectNames;

		Array<PhaseGroup*> theGroups;
		
	private:
		ErrorStat theErrorStatCache;
};

ErrorStrTbl PhaseGroup::TheErrStrs;

static MyOpts TheOpts;
static strstream ReportBuf;

/* MyOpt */

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "<log_file_name> ...";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	for (int i = 0 ; i < opts.count(); ++i)
		theFiles.append(new String(opts[i]));

	// add default fname if none are specified
	if (!theFiles.count())
		theFiles.append(new String("-"));

	return theFiles.count() > 0;
}

/* PhaseGroup */

PhaseGroup::PhaseGroup(const String &aName): name(aName) {
}

/* MyScanner */

MyScanner::MyScanner(Strs phaseNames, Strs objectNames):
	thePhaseNames(phaseNames), theObjectNames(objectNames) {

	for (int i = 0; i < thePhaseNames.count(); ++i)
		theGroups.append(new PhaseGroup(*thePhaseNames[i]));
}

MyScanner::~MyScanner() {
	while (theGroups.count()) delete theGroups.pop();
}

bool MyScanner::load(ILog &il, const LogEntryPx &px) {
	switch (px.theTag) {
		case lgErrorStrs: {
			// should be called only once per log
			ErrorStrTbl est;
			est.load(il);
			PhaseGroup::TheErrStrs = est;
			break;
		}
		case lgErrorStat: {
			ErrorStat es;
			es.load(il);
			theErrorStatCache = es;
			break;
		}
		case lgStatPhaseRec: {
			StatPhaseRec ph;
			ph.load(il);
			notePhase(ph);
			break;
		}
	}
	return true;
}

void MyScanner::notePhase(const StatPhaseRec &ph) {
	// check if we need this phase
	bool found = !thePhaseNames.count();
	for (int i = 0; !found && i < thePhaseNames.count(); ++i)
		found = ph.name()== *thePhaseNames[i];
	if (!found)
		return;

	// check if we saw this phase before
	PhaseGroup *group = 0;
	for (int i = 0; !group && i < theGroups.count(); ++i) {
		if (theGroups[i]->name == ph.name())
			group = theGroups[i];
	}

	if (!group) {
		Assert(!thePhaseNames.count());
		group = new PhaseGroup(ph.name());
		theGroups.append(group);
	}

	group->errors.add(theErrorStatCache);
	group->rec.merge(ph);
}

void MyScanner::report() const {
	if (!theGroups.count()) {
		cerr << "no matching phases found" << endl;
		exit(-1);
	}

	// concatenate all phases
	PhaseGroup group(0);
	for (int i = 0; i < theGroups.count(); ++i) {
		const PhaseGroup &g = *theGroups[i];
		group.errors.add(g.errors);
		group.rec.concat(g.rec);
	}
	
	group.rec.print(ReportBuf, "");
	group.errors.print(ReportBuf, PhaseGroup::TheErrStrs, "errors.");

	ReportBuf << "log.count:\t " << TheOpts.theFiles.count() << endl;

	ReportBuf << ends;
	if (MatchAndPrint(cout, ReportBuf.str(), theObjectNames) == 0) {
		cerr << "no logged objects matched any of the requested object names" << endl;
		exit(-2);
	}
}


/* local routines */

static
void configureLogs(int prec) {
	if (TheOpts.theOutFileName && TheOpts.theOutFileName != "-")
		cout = *new ofstream(TheOpts.theOutFileName);

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);

	configureStream(ReportBuf, prec);
}

static
void configure() {
	configureLogs(2);
}

int main(int argc, char *argv[]) {

	CmdLine cmd;
	cmd.configure(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	MyScanner scanner(TheOpts.thePhases.val(), TheOpts.theObjects.val());
	for (int i = 0; i < TheOpts.theFiles.count(); ++i) {
		const String &fname = *TheOpts.theFiles[i];
		ILog log;
		if (fname == "-")
			log.stream("stdin", &cin);
		else
			log.stream(fname, (istream*)new ifstream(fname));

		for (LogIter li(&log); li; ++li)
			scanner.load(log, *li);
	}

	scanner.report();

	return 0;
}
